/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <string>
#include <sstream>
#include <list>

#include "point.h"
#include "spakes.h"

#ifdef WINDOWS
class Window;
#else
class Gameboard;
#endif
class Event;
class UDPSocket;

class GameField
{
public:
	GameField();
	void add_dot(const Point &);
	void remove_dot(const Point &);
	bool free(const Point &);
	bool inside(const Point &);
	char fieldval(Point &p) {
		if (!inside(p)) return 2;
		return field[p.x][p.y];
	}
private:
	char field[BOARDXSIZE][BOARDYSIZE];
};


class Player
{
public:
	int Score() { return score; }
	std::string Name() { return name; }
	bool outside();
	virtual void process_event(const Event &) { }
	virtual void lose_life();
	bool collision();
	void process_event(const std::string &);

	void draw(bool fullredraw = false);
	void undraw_begin();
	void undraw();
	void announce();

	int num_ticks() const { return numticks; }
	bool Living() const { return ((numlives > 0) && ingame); }
	bool InGame() const { return ingame; }
#ifdef WINDOWS
	Player(Window *, int, std::istringstream &);
#else
	Player(Gameboard *, int, std::istringstream &);
#endif
	virtual ~Player();

	virtual void tick();
	virtual void send_tick(Player *) { }
#ifdef WINDOWS
	friend class Window;
#else
	friend class Gameboard;
#endif
	friend class Server;
	friend bool PlayerCollision(int );
	friend bool FreePos(Point &);
	friend class RemotePrimaryPlayer;

protected:
	std::list<Point> dots;
	int score, playerid, numticks, numlives;
	Point currpos;
#ifdef WINDOWS
	Window *w;
#else
	Gameboard *w;
#endif
	std::string name;
	bool ingame;
};

// nodes for A* search

struct ASNode
{
	int fitness;
	Point pos;
	std::list<Point> path;
	ASNode(Point p) : pos(p) { }
};

typedef std::list<ASNode> nodes;
typedef nodes::iterator nodes_it;

class AIPlayer : public Player
{
public:
#ifdef WINDOWS
	AIPlayer(Window *, int, std::istringstream &);
#else
	AIPlayer(Gameboard *, int, std::istringstream &);
#endif

	void tick();
private:
	bool choose_target();
	void push_successors(Point &, ASNode);
	bool compute_route(Point &, Point &);
	nodes openlist, closedlist;
	std::list<Point> path;
	void clear_lists();
	void lose_life();
	Point headingfor, lastdirection, startingpos;
};


class Consoleplayer : public Player
{
public:
	void tick();
#ifdef WINDOWS
	Consoleplayer(Window *, int , std::istringstream &, bool pusearrows = true);
#else
	Consoleplayer(Gameboard *, int , std::istringstream &, bool pusearrows = true);
#endif
	void lose_life();
	void process_event(const Event &);

protected:
	int direction, startingdirection;
	Point startingpos;
	bool usearrows, keypressed[4];
};


// The remote "primary" player receives the ticks of our "primary" and
// "secondary" players. We could also define the class "RemoteSecondaryPlayer",
// but we use "Player" directly instead. The remote "secondary" player is not
// sent any ticks.

class RemotePrimaryPlayer : public Player
{
public:
#ifdef WINDOWS
	RemotePrimaryPlayer(Window *, int, std::istringstream &);
#else
	RemotePrimaryPlayer(Gameboard *, int, std::istringstream &);
#endif
	~RemotePrimaryPlayer();
	void send_tick(Player *);

private:
	std::ostringstream *scheduled_ticks;
	int num_scheduled_ticks;
	UDPSocket *socket;
};


